# Janos version 1.2.1
# JanosPrinter.py
# by Antoine Rauzy
# Copyrights (c) 2018-2022 AltaRica Association
#
# This file is part of Janos, a pedagogical stochastic simulator.
# Janos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Janos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Janos.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Models
#   3.1) Models
#   3.2) Blocks and classes
#   3.3) Variables and equations

# 1) Imported modules
# -------------------

from S2MLCore import *

from JanosModels import *
from JanosToken import *

# 2) Creator
# ----------

class JanosModelPrinter(S2MLModelPrinter):
  def __init__(self, engine):
    S2MLModelPrinter.__init__(self, engine)
    self._location = JanosType.UNKNOWN

# 3) Models
# ---------

# 3.1) Models
# -----------

  def PrintElement(self, element, output):
    if element.GetType()==S2MLType.DOMAIN_DECLARATION:
      self._UpdateLocation(JanosType.UNKNOWN, output)
      self.PrintDomainDeclaration(element, output)
    elif element.GetType()==JanosType.BLOCK:
      self._UpdateLocation(JanosType.UNKNOWN, output)
      self.PrintBlock(element, output)
    elif element.GetType()==JanosType.CLASS:
      self._UpdateLocation(JanosType.UNKNOWN, output)
      self.PrintClass(element, output)
    elif element.GetType()==JanosType.INSTANCE:
      self._UpdateLocation(JanosType.UNKNOWN, output)
      self.PrintInstance(element, output)
    elif element.GetType()==JanosType.CLONES:
      self._UpdateLocation(JanosType.UNKNOWN, output)
      self.PrintClonesDirective(element, output)
    elif element.GetType()==JanosType.EXTENDS:
      self._UpdateLocation(JanosType.UNKNOWN, output)
      self.PrintExtendsDirective(element, output)
    elif element.GetType()==JanosType.EMBEDS:
      self._UpdateLocation(JanosType.UNKNOWN, output)
      self.PrintEmbedsDirective(element, output)
    elif element.GetType()==JanosType.PARAMETER:
      self._UpdateLocation(JanosType.UNKNOWN, output)
      self.PrintParameter(element, output)
    elif element.GetType()==JanosType.VARIABLE:
      self._UpdateLocation(JanosType.UNKNOWN, output)
      self.PrintVariable(element, output)
    elif element.GetType()==JanosType.ASSIGNMENT:
      self._UpdateLocation(JanosType.EQUATION, output)
      self.PrintAssignment(element, output)
    elif element.GetType()==JanosType.DOUBLE_ASSIGNMENT:
      self._UpdateLocation(JanosType.EQUATION, output)
      self.PrintDoubleAssignment(element, output)
    elif element.GetType()==JanosType.OBSERVER:
      self._UpdateLocation(JanosType.UNKNOWN, output)
      self.PrintObserver(element, output)
    elif element.GetType()==S2MLType.FUNCTION:
      self._UpdateLocation(JanosType.UNKNOWN, output)
      self.PrintFunction(element, output)
    elif element.GetType()==S2MLType.FUNCTION_CALL:
      self._UpdateLocation(JanosType.UNKNOWN, output)
      self.PrintFunctionCall(element, output)
    elif element.GetType()==S2MLType.OPERATOR:
      self._UpdateLocation(JanosType.UNKNOWN, output)
      self.PrintOperator(element, output)

  def _UpdateLocation(self, location, output):
    if self._location!=JanosType.UNKNOWN and location==JanosType.UNKNOWN:
      self._location = JanosType.UNKNOWN
      self._depth -= 1
    elif self._location!=JanosType.EQUATION and location==JanosType.EQUATION:
      self._location = JanosType.EQUATION
      self.PrintIndentation(output)
      self.PrintKeyword(JanosToken.KEYWORD_ASSERTION, output)
      output.write("\n")
      self._depth += 1

# 3.2) Blocks and classes
# -----------------------

  def _PrintBlockBody(self, block, output):
    self._location = JanosType.UNKNOWN
    self._depth += 1
    for element in block.GetElements():
      self.PrintElement(element, output)
    self._UpdateLocation(JanosType.UNKNOWN, output)
    self._depth -= 1

# 3.3) Variables and equations
# ----------------------------

  def PrintVariable(self, variable, output):
    self.PrintIndentation(output)
    self.PrintDomain(variable.GetDomain(), output)
    output.write(" ")
    self.PrintPath(variable.GetPath(), output)
    self.PrintAttributes(variable, output)
    output.write(";\n")

  def PrintAssignment(self, assignment, output):
    self.PrintIndentation(output)
    self.PrintStochasticExpression(assignment.GetLeftMember(), output)
    output.write(" := ")
    self.PrintStochasticExpression(assignment.GetRightMember(), output)
    output.write(";\n")

  def PrintDoubleAssignment(self, assignment, output):
    self.PrintIndentation(output)
    self.PrintStochasticExpression(assignment.GetLeftMember(), output)
    output.write(" :=: ")
    self.PrintStochasticExpression(assignment.GetRightMember(), output)
    output.write(";\n")

# 3.4) Observers
# --------------

  def PrintObserver(self, observer, output):
    self._UpdateLocation(False, output)
    self.PrintIndentation(output)
    self.PrintKeyword(JanosToken.KEYWORD_OBSERVER, output)
    output.write(" ")
    self.PrintDomain(observer.GetDomain(), output)
    output.write(" ")
    self.PrintPath(observer.GetPath(), output)
    output.write(" = ")
    self.PrintStochasticExpression(observer.GetDefinition(), output)
    output.write(";\n")
