# S2MLCore version 1.1.1
# S2MLStochasticExpressionManager.py
# by Antoine Rauzy
# Copyrights (c) 2018-2022 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Management of operators
# 4) Calculation of values of expressions
# 5) Management of random deviates

# 1) Imported modules
# -------------------

import string
import sys
import re
import math
import random

from S2MLType import *
from S2MLToken import *
from S2MLMetaModel import *
from S2MLDomains import *
from S2MLStochasticExpression import *
from S2MLStochasticOperator import *

# 2) Creator
# ----------

class S2MLStochasticExpressionManager:

  def __init__(self, engine):
    from S2MLEngine import S2MLEngine
    self._engine = engine
    self._operatorsByTypes = dict()
    self._operatorsByKeywordCodes = dict()
    self.RegisterOperators()
    self._missionTime = 0.0
    self._calculateValueOfReferenceExpression = S2MLStochasticExpressionManager.CalculateValueOfReferenceExpression
    self._calculateValueOfElement = S2MLStochasticExpressionManager.CalculateValueOfElement

  def RegisterOperators(self):
    self.RegisterOperator(S2MLType.ADD_EXPRESSION, S2MLToken.TOKEN_PLUS, 1, S2MLStochasticOperator.MANY, S2MLStochasticExpressionManager.CalculateAdd)
    self.RegisterOperator(S2MLType.SUB_EXPRESSION, S2MLToken.TOKEN_MINUS, 1, S2MLStochasticOperator.MANY, S2MLStochasticExpressionManager.CalculateSub)
    self.RegisterOperator(S2MLType.MUL_EXPRESSION, S2MLToken.TOKEN_STAR, 1, S2MLStochasticOperator.MANY, S2MLStochasticExpressionManager.CalculateMul)
    self.RegisterOperator(S2MLType.DIV_EXPRESSION, S2MLToken.TOKEN_SLASH, 1, S2MLStochasticOperator.MANY, S2MLStochasticExpressionManager.CalculateDiv)
    self.RegisterOperator(S2MLType.NEG_EXPRESSION, S2MLToken.TOKEN_NEG, 1, 1, S2MLStochasticExpressionManager.CalculateNeg)
    self.RegisterOperator(S2MLType.MIN_EXPRESSION, S2MLToken.KEYWORD_MIN, 1, S2MLStochasticOperator.MANY, S2MLStochasticExpressionManager.CalculateMin)
    self.RegisterOperator(S2MLType.MAX_EXPRESSION, S2MLToken.KEYWORD_MAX, 1, S2MLStochasticOperator.MANY, S2MLStochasticExpressionManager.CalculateMax)
    self.RegisterOperator(S2MLType.AND_EXPRESSION, S2MLToken.KEYWORD_AND, 1, S2MLStochasticOperator.MANY, S2MLStochasticExpressionManager.CalculateAnd)
    self.RegisterOperator(S2MLType.OR_EXPRESSION, S2MLToken.KEYWORD_OR, 1, S2MLStochasticOperator.MANY, S2MLStochasticExpressionManager.CalculateOr)
    self.RegisterOperator(S2MLType.NOT_EXPRESSION, S2MLToken.KEYWORD_NOT, 1, 1, S2MLStochasticExpressionManager.CalculateNot)
    self.RegisterOperator(S2MLType.EQ_EXPRESSION, S2MLToken.TOKEN_EQ, 2, 2, S2MLStochasticExpressionManager.CalculateEq)
    self.RegisterOperator(S2MLType.DF_EXPRESSION, S2MLToken.TOKEN_DF, 2, 2, S2MLStochasticExpressionManager.CalculateDf)
    self.RegisterOperator(S2MLType.LT_EXPRESSION, S2MLToken.TOKEN_LT, 2, 2, S2MLStochasticExpressionManager.CalculateLt)
    self.RegisterOperator(S2MLType.GT_EXPRESSION, S2MLToken.TOKEN_GT, 2, 2, S2MLStochasticExpressionManager.CalculateGt)
    self.RegisterOperator(S2MLType.LEQ_EXPRESSION, S2MLToken.TOKEN_LEQ, 2, 2, S2MLStochasticExpressionManager.CalculateLeq)
    self.RegisterOperator(S2MLType.GEQ_EXPRESSION, S2MLToken.TOKEN_GEQ, 2, 2, S2MLStochasticExpressionManager.CalculateGeq)
    self.RegisterOperator(S2MLType.WITHIN_EXPRESSION, S2MLToken.KEYWORD_WITHIN, 1, 1, S2MLStochasticExpressionManager.CalculateWithin)
    self.RegisterOperator(S2MLType.COUNT_EXPRESSION, S2MLToken.KEYWORD_COUNT, 1, S2MLStochasticOperator.MANY, S2MLStochasticExpressionManager.CalculateCount)
    self.RegisterOperator(S2MLType.IF_EXPRESSION, S2MLToken.KEYWORD_IF, 3, 3, S2MLStochasticExpressionManager.CalculateIf)
    self.RegisterOperator(S2MLType.UNIFORM_DEVIATE_EXPRESSION, S2MLToken.KEYWORD_UNIFORM_DEVIATE, 2, 2, S2MLStochasticExpressionManager.CalculateUniformDeviate)
    self.RegisterOperator(S2MLType.NORMAL_DEVIATE_EXPRESSION, S2MLToken.KEYWORD_NORMAL_DEVIATE, 2, 2, S2MLStochasticExpressionManager.CalculateNormalDeviate)
    self.RegisterOperator(S2MLType.LOGNORMAL_DEVIATE_EXPRESSION, S2MLToken.KEYWORD_LOGNORMAL_DEVIATE, 2, 2, S2MLStochasticExpressionManager.CalculateLognormalDeviate)
    self.RegisterOperator(S2MLType.TRIANGULAR_DEVIATE_EXPRESSION, S2MLToken.KEYWORD_TRIANGULAR_DEVIATE, 3, 3, S2MLStochasticExpressionManager.CalculateTriangularDeviate)
    self.RegisterOperator(S2MLType.EXPONENTIAL_DEVIATE_EXPRESSION, S2MLToken.KEYWORD_EXPONENTIAL_DEVIATE, 1, 1, S2MLStochasticExpressionManager.CalculateExponentialDeviate)
    self.RegisterOperator(S2MLType.WEIBULL_DEVIATE_EXPRESSION, S2MLToken.KEYWORD_WEIBULL_DEVIATE, 2, 2, S2MLStochasticExpressionManager.CalculateWeibullDeviate)
    self.RegisterOperator(S2MLType.PIECEWISE_UNIFORM_DEVIATE_EXPRESSION, S2MLToken.KEYWORD_PIECEWISE_UNIFORM_DEVIATE, 0, 0, S2MLStochasticExpressionManager.CalculatePiecewiseUniformDeviate)
    self.RegisterOperator(S2MLType.STEPWISE_DEVIATE_EXPRESSION, S2MLToken.KEYWORD_STEPWISE_DEVIATE, 0, 0, S2MLStochasticExpressionManager.CalculateStepwiseDeviate)
    self.RegisterOperator(S2MLType.MISSION_TIME_EXPRESSION, S2MLToken.KEYWORD_MISSION_TIME, 0, 0, S2MLStochasticExpressionManager.CalculateMissionTime)
    self.RegisterOperator(S2MLType.EXP_EXPRESSION, S2MLToken.KEYWORD_EXP, 1, 1, S2MLStochasticExpressionManager.CalculateExp)
    self.RegisterOperator(S2MLType.LOG_EXPRESSION, S2MLToken.KEYWORD_LOG, 1, 1, S2MLStochasticExpressionManager.CalculateLog)
    self.RegisterOperator(S2MLType.POW_EXPRESSION, S2MLToken.KEYWORD_POW, 2, 2, S2MLStochasticExpressionManager.CalculatePow)
    self.RegisterOperator(S2MLType.SQRT_EXPRESSION, S2MLToken.KEYWORD_SQRT, 1, 1, S2MLStochasticExpressionManager.CalculateSqrt)
    self.RegisterOperator(S2MLType.FLOOR_EXPRESSION, S2MLToken.KEYWORD_FLOOR, 1, 1, S2MLStochasticExpressionManager.CalculateFloor)
    self.RegisterOperator(S2MLType.CEIL_EXPRESSION, S2MLToken.KEYWORD_CEIL, 1, 1, S2MLStochasticExpressionManager.CalculateCeil)
    self.RegisterOperator(S2MLType.ABS_EXPRESSION, S2MLToken.KEYWORD_ABS, 1, 1, S2MLStochasticExpressionManager.CalculateAbs)
    self.RegisterOperator(S2MLType.MOD_EXPRESSION, S2MLToken.KEYWORD_MOD, 2, 2, S2MLStochasticExpressionManager.CalculateMod)
    self.RegisterOperator(S2MLType.GAMMA_EXPRESSION, S2MLToken.KEYWORD_GAMMA, 1, 1, S2MLStochasticExpressionManager.CalculateGamma)
    self.RegisterOperator(S2MLType.SIN_EXPRESSION, S2MLToken.KEYWORD_SIN, 1, 1, S2MLStochasticExpressionManager.CalculateSin)
    self.RegisterOperator(S2MLType.COS_EXPRESSION, S2MLToken.KEYWORD_COS, 1, 1, S2MLStochasticExpressionManager.CalculateCos)
    self.RegisterOperator(S2MLType.TAN_EXPRESSION, S2MLToken.KEYWORD_TAN, 1, 1, S2MLStochasticExpressionManager.CalculateTan)
    self.RegisterOperator(S2MLType.ASIN_EXPRESSION, S2MLToken.KEYWORD_ASIN, 1, 1, S2MLStochasticExpressionManager.CalculateAsin)
    self.RegisterOperator(S2MLType.ACOS_EXPRESSION, S2MLToken.KEYWORD_ACOS, 1, 1, S2MLStochasticExpressionManager.CalculateAcos)
    self.RegisterOperator(S2MLType.ATAN_EXPRESSION, S2MLToken.KEYWORD_ATAN, 1, 1, S2MLStochasticExpressionManager.CalculateAtan)
    self.RegisterOperator(S2MLType.CONSTANT_DISTRIBUTION_EXPRESSION, S2MLToken.KEYWORD_CONSTANT_DISTRIBUTION, 1, 1, S2MLStochasticExpressionManager.CalculateConstantDistribution)
    self.RegisterOperator(S2MLType.EXPONENTIAL_DISTRIBUTION_EXPRESSION, S2MLToken.KEYWORD_EXPONENTIAL_DISTRIBUTION, 2, 2, S2MLStochasticExpressionManager.CalculateExponentialDistribution)
    self.RegisterOperator(S2MLType.GLM_DISTRIBUTION_EXPRESSION, S2MLToken.KEYWORD_GLM_DISTRIBUTION, 4, 4, S2MLStochasticExpressionManager.CalculateGLMDistribution)
    self.RegisterOperator(S2MLType.WEIBULL_DISTRIBUTION_EXPRESSION, S2MLToken.KEYWORD_WEIBULL_DISTRIBUTION, 3, 3, S2MLStochasticExpressionManager.CalculateWeibullDistribution)
    self.RegisterOperator(S2MLType.DIRAC_DISTRIBUTION_EXPRESSION, S2MLToken.KEYWORD_DIRAC_DISTRIBUTION, 2, 2, S2MLStochasticExpressionManager.CalculateDiracDistribution)
    self.RegisterOperator(S2MLType.PIECEWISE_UNIFORM_DISTRIBUTION_EXPRESSION, S2MLToken.KEYWORD_PIECEWISE_UNIFORM_DISTRIBUTION, 1, 1, S2MLStochasticExpressionManager.CalculatePiecewiseUniformDistribution)
    self.RegisterOperator(S2MLType.STEPWISE_DISTRIBUTION_EXPRESSION, S2MLToken.KEYWORD_STEPWISE_DISTRIBUTION, 1, 1, S2MLStochasticExpressionManager.CalculateStepwiseDistribution)

  def GetEngine(self):
    return self._engine

  def GetMissionTime(self):
    return self._missionTime
  def SetMissionTime(self, missionTime):
    self._missionTime = missionTime

  def GetCalculateValueOfReferenceExpression(self):
    return self._calculateValueOfReferenceExpression

  def SetCalculateValueOfReferenceExpression(self, calculateValueOfReferenceExpression):
    self._calculateValueOfReferenceExpression = calculateValueOfReferenceExpression

  def SetCalculateValueOfElement(self, calculateValueOfElement):
    self._calculateValueOfElement = calculateValueOfElement

  def GetCalculateValueOfElement(self):
    return self._calculateValueOfElement

  def SetCalculateValueOfElement(self, calculateValueOfElement):
    self._calculateValueOfElement = calculateValueOfElement

# 3) Management of operators
# --------------------------

  def GetOperatorOfType(self, type):
    return self._operatorsByTypes.get(type, None)

  def GetOperatorOfKeywordCode(self, keywordCode):
    return self._operatorsByKeywordCodes.get(keywordCode, None)

  def RegisterOperator(self, type, keywordCode, minimumNumberOfArguments, maximumNumberOfArguments, calculate):
    if self.GetOperatorOfType(type)!=None or self.GetOperatorOfKeywordCode(keywordCode)!=None:
      return
    operator = S2MLStochasticOperator(type, keywordCode, minimumNumberOfArguments, maximumNumberOfArguments, calculate)
    self._operatorsByTypes[type] = operator
    self._operatorsByKeywordCodes[keywordCode] = operator

  def GetKeywordCodeOfOperatorOfType(self, type):
    operator = self.GetOperatorOfType(type)
    if operator==None:
      return 0
    return operator._keywordCode

  # def IsRandomDeviate(self, type):
    # operator = self.GetOperatorOfType(type)
    # if operator==None:
      # return False
    # return operator.IsRandomDeviate()

# 4) Calculation of values of expressions
# ---------------------------------------

  def ResetValuesOfParameters(self, container):
    for element in container.GetElements():
      if element.GetType()==S2MLType.BLOCK:
        self.ResetValuesOfParameters(element)
      elif element.GetType()==S2MLType.PARAMETER:
        element.ResetValue()

  def CalculateValueOfParameter(self, parameter):
    if parameter.IsValueCalculated():
      value = parameter.GetValue()
    else:
      definition = parameter.GetDefinition()
      value = self.CalculateValueOfExpression(definition)
      parameter.SetValue(value)
    # print("parameter(" + parameter.GetIdentifier() + ") = " + str(value.ToFloat()))
    return value

  @staticmethod
  def CalculateValueOfReferenceExpression(expressionManager, expression):
    referredElement = expression.GetReferredElement()
    if referredElement==None:
      return S2MLConstant.NewNone()
    value = expressionManager._calculateValueOfElement(expressionManager, referredElement)
    expression.SetValue(value)
    return expression.GetValue()

  @staticmethod
  def CalculateValueOfElement(expressionManager, element):
    if element.GetType()==S2MLType.PARAMETER:
      return expressionManager.CalculateValueOfParameter(element)
    return S2MLConstant.NewNone()

  def CalculateValueOfExpression(self, expression):
    if expression.GetType()==S2MLType.REFERENCE_EXPRESSION:
      return self._calculateValueOfReferenceExpression(self, expression)
    if expression.GetType()==S2MLType.CONSTANT_EXPRESSION:
      return expression.GetValue()
    operator = self.GetOperatorOfType(expression.GetType())
    if operator==None:
      return S2MLConstant.NewNone()
    return operator._calculate(self, expression)

  def CalculateAdd(self, expression):
    value = expression.GetValue()
    value.SetAsInteger(0)
    for child in expression.GetChildren():
      childValue = self.CalculateValueOfExpression(child)
      value.AddMe(childValue)
    return value

  def CalculateSub(self, expression):
    value = expression.GetValue()
    value.SetAsInteger(0)
    first = True
    for child in expression.GetChildren():
      childValue = self.CalculateValueOfExpression(child)
      if first:
        first = False
        value.Set(childValue)
      else:
        value.SubMe(childValue)
    return value

  def CalculateMul(self, expression):
    value = expression.GetValue()
    value.SetAsInteger(1)
    for child in expression.GetChildren():
      childValue = self.CalculateValueOfExpression(child)
      value.MulMe(childValue)
    return value

  def CalculateDiv(self, expression):
    value = expression.GetValue()
    value.SetAsInteger(0)
    first = True
    for child in expression.GetChildren():
      childValue = self.CalculateValueOfExpression(child)
      if first:
        first = False
        value.Set(childValue)
      else:
        value.DivMe(childValue)
    return value

  def CalculateNeg(self, expression):
    value = expression.GetValue()
    child = expression.GetChild(0)
    childValue = self.CalculateValueOfExpression(child)
    value.NegMe(childValue)
    return value

  def CalculateMin(self, expression):
    value = expression.GetValue()
    value.SetAsPosInf()
    first = True
    for child in expression.GetChildren():
      childValue = self.CalculateValueOfExpression(child)
      value.MinMe(childValue)
    return value

  def CalculateMax(self, expression):
    value = expression.GetValue()
    value.SetAsNegInf()
    first = True
    for child in expression.GetChildren():
      childValue = self.CalculateValueOfExpression(child)
      value.MaxMe(childValue)
    return value

  def CalculateAnd(self, expression):
    value = expression.GetValue()
    value.SetAsBoolean(True)
    for child in expression.GetChildren():
      childValue = self.CalculateValueOfExpression(child)
      value.AndMe(childValue)
    return value

  def CalculateOr(self, expression):
    value = expression.GetValue()
    value.SetAsBoolean(False)
    for child in expression.GetChildren():
      childValue = self.CalculateValueOfExpression(child)
      value.OrMe(childValue)
    return value

  def CalculateNot(self, expression):
    value = expression.GetValue()
    value.SetAsBoolean(False)
    child = expression.GetChild(0)
    childValue = self.CalculateValueOfExpression(child)
    value.NotMe(childValue)
    return value

  def CalculateEq(self, expression):
    value = expression.GetValue()
    value.SetAsBoolean(False)
    child1 = expression.GetChild(0)
    childValue1 = self.CalculateValueOfExpression(child1)
    child2 = expression.GetChild(1)
    childValue2 = self.CalculateValueOfExpression(child2)
    value.EqMe(childValue1, childValue2)
    return value

  def CalculateDf(self, expression):
    value = expression.GetValue()
    value.SetAsBoolean(False)
    child1 = expression.GetChild(0)
    childValue1 = self.CalculateValueOfExpression(child1)
    child2 = expression.GetChild(1)
    childValue2 = self.CalculateValueOfExpression(child2)
    value.DfMe(childValue1, childValue2)
    return value

  def CalculateLt(self, expression):
    value = expression.GetValue()
    value.SetAsBoolean(False)
    child1 = expression.GetChild(0)
    childValue1 = self.CalculateValueOfExpression(child1)
    child2 = expression.GetChild(1)
    childValue2 = self.CalculateValueOfExpression(child2)
    value.LtMe(childValue1, childValue2)
    return value

  def CalculateGt(self, expression):
    value = expression.GetValue()
    value.SetAsBoolean(False)
    child1 = expression.GetChild(0)
    childValue1 = self.CalculateValueOfExpression(child1)
    child2 = expression.GetChild(1)
    childValue2 = self.CalculateValueOfExpression(child2)
    value.GtMe(childValue1, childValue2)
    return value

  def CalculateLeq(self, expression):
    value = expression.GetValue()
    value.SetAsBoolean(False)
    child1 = expression.GetChild(0)
    childValue1 = self.CalculateValueOfExpression(child1)
    child2 = expression.GetChild(1)
    childValue2 = self.CalculateValueOfExpression(child2)
    value.LeqMe(childValue1, childValue2)
    return value

  def CalculateGeq(self, expression):
    value = expression.GetValue()
    value.SetAsBoolean(False)
    child1 = expression.GetChild(0)
    childValue1 = self.CalculateValueOfExpression(child1)
    child2 = expression.GetChild(1)
    childValue2 = self.CalculateValueOfExpression(child2)
    value.GeqMe(childValue1, childValue2)
    return value

  def CalculateWithin(self, expression):
    value = expression.GetValue()
    value.SetAsBoolean(False)
    child = expression.GetChild(0)
    childValue = self.CalculateValueOfExpression(child)
    domain = expression.GetDomain(1)
    for constant in domain.GetConstants():
      if S2MLConstant.IsEq(constant, childValue):
        value.SetAsBoolean(True)
        break
    return value

  def CalculateCount(self, expression):
    value = expression.GetValue()
    value.SetAsInteger(0)
    for child in expression.GetChildren():
      childValue = self.CalculateValueOfExpression(child)
      value.CountMe(childValue)
    return value

  def CalculateIf(self, expression):
    value = expression.GetValue()
    child1 = expression.GetChild(0)
    childValue1 = self.CalculateValueOfExpression(child1)
    if childValue1.IsBoolean():
      if childValue1.GetValue():
        child2 = expression.GetChild(1)
      else:
        child2 = expression.GetChild(2)
      childValue2 = self.CalculateValueOfExpression(child2)
      value.Set(childValue2)
    else:
      value.Reset()
    return value

  def CalculateUniformDeviate(self, expression):
    value = expression.GetValue()
    lowChild = expression.GetChild(0)
    lowChildValue = self.CalculateValueOfExpression(lowChild)
    if not lowChildValue.IsNumber():
      value.Reset()
      return value
    highChild = expression.GetChild(1)
    highChildValue = self.CalculateValueOfExpression(highChild)
    if not highChildValue.IsNumber():
      value.Reset()
      return value
    low = lowChildValue.GetValue()
    high = highChildValue.GetValue()
    value.SetAsReal(low + expression.GetRandomDeviateAbscissa()*(high-low))
    return value

  def CalculateNormalDeviate(self, expression):
    value = expression.GetValue()
    muChild = expression.GetChild(0)
    muChildValue = self.CalculateValueOfExpression(muChild)
    if not muChildValue.IsNumber():
      value.Reset()
      return value
    sigmaChild = expression.GetChild(1)
    sigmaChildValue = self.CalculateValueOfExpression(sigmaChild)
    if not sigmaChildValue.IsNumber():
      value.Reset()
      return value
    mu = muChildValue.GetValue()
    sigma = sigmaChildValue.GetValue()
    value.SetAsReal(mu + sigma*expression.GetRandomDeviateAbscissa())
    return value

  def CalculateLognormalDeviate(self, expression):
    value = expression.GetValue()
    muChild = expression.GetChild(0)
    muChildValue = self.CalculateValueOfExpression(muChild)
    if not muChildValue.IsNumber():
      value.Reset()
      return value
    sigmaChild = expression.GetChild(1)
    sigmaChildValue = self.CalculateValueOfExpression(sigmaChild)
    if not sigmaChildValue.IsNumber():
      value.Reset()
      return value
    mu = muChildValue.GetValue()
    sigma = sigmaChildValue.GetValue()
    value.SetAsReal(math.exp(mu + sigma*expression.GetRandomDeviateAbscissa()))
    return value

  def CalculateTriangularDeviate(self, expression):
    value = expression.GetValue()
    lowChild = expression.GetChild(0)
    lowChildValue = self.CalculateValueOfExpression(lowChild)
    if not lowChildValue.IsNumber():
      value.Reset()
      return value
    highChild = expression.GetChild(1)
    highChildValue = self.CalculateValueOfExpression(highChild)
    if not highChildValue.IsNumber():
      value.Reset()
      return value
    modeChild = expression.GetChild(2)
    modeChildValue = self.CalculateValueOfExpression(modeChild)
    low = lowChildValue.GetValue()
    high = highChildValue.GetValue()
    mode = modeChildValue.GetValue()
    f = (mode-low)/(high-low)
    U = expression.GetRandomDeviateAbscissa()
    if U<f:
      value.SetAsReal(low+math.sqrt(U*(high-low)*(mode-low)))
    else:
      value.SetAsReal(high-math.sqrt((1-U)*(high-low)*(high-mode)))
    return value

  def CalculateExponentialDeviate(self, expression):
    value = expression.GetValue()
    rateChild = expression.GetChild(0)
    rateChildValue = self.CalculateValueOfExpression(rateChild)
    if not rateChildValue.IsNumber():
      value.Reset()
      return value
    rate = rateChildValue.GetValue()
    if rate==0.0:
      value.SetAsPosInf()
    else:
      delay = -math.log(expression.GetRandomDeviateAbscissa())/rate
      value.SetAsReal(delay)
    return value

  def CalculateWeibullDeviate(self, expression):
    value = expression.GetValue()
    shapeChild = expression.GetChild(0)
    shapeChildValue = self.CalculateValueOfExpression(shapeChild)
    if not shapeChildValue.IsNumber():
      value.Reset()
      return value
    scaleChild = expression.GetChild(1)
    scaleChildValue = self.CalculateValueOfExpression(scaleChild)
    if not scaleChildValue.IsNumber():
      value.Reset()
      return value
    shape = shapeChildValue.GetValue()
    scale = scaleChildValue.GetValue()
    if scale==0.0:
      value.SetAsPosInf()
    else:
      delay = shape*math.pow(-math.log(expression.GetRandomDeviateAbscissa()), 1/scale)
      value.SetAsReal(delay)
    return value

  def CalculatePiecewiseUniformDeviate(self, expression):
    value = expression.GetValue()
    distribution = expression.GetDistribution()
    delay = distribution.PickUpXInterpolated(expression.GetRandomDeviateAbscissa())
    value.SetAsReal(delay)
    return value

  def CalculateStepwiseDeviate(self, expression):
    value = expression.GetValue()
    distribution = expression.GetDistribution()
    delay = distribution.PickUpXStepwise(expression.GetRandomDeviateAbscissa())
    value.SetAsReal(delay)
    return value

  def CalculateMissionTime(self, expression):
    value = expression.GetValue()
    value.SetAsReal(self._missionTime)
    return value

  def CalculateExp(self, expression):
    value = expression.GetValue()
    child = expression.GetChild(0)
    childValue = self.CalculateValueOfExpression(child)
    if not childValue.IsNumber():
      value.Reset()
      return value
    value.SetAsReal(math.exp(childValue.GetValue()))
    return value

  def CalculateLog(self, expression):
    value = expression.GetValue()
    child = expression.GetChild(0)
    childValue = self.CalculateValueOfExpression(child)
    if not childValue.IsNumber():
      value.Reset()
      return value
    value.SetAsReal(math.log(childValue.GetValue()))
    return value

  def CalculatePow(self, expression):
    value = expression.GetValue()
    child1 = expression.GetChild(0)
    childValue1 = self.CalculateValueOfExpression(child1)
    child2 = expression.GetChild(1)
    childValue2 = self.CalculateValueOfExpression(child2)
    if not childValue1.IsNumber() or not childValue2.IsNumber():
      value.Reset()
      return value
    value.SetAsReal(math.pow(childValue1.GetValue(), childValue2.GetValue()))
    return value

  def CalculateSqrt(self, expression):
    value = expression.GetValue()
    child = expression.GetChild(0)
    childValue = self.CalculateValueOfExpression(child)
    if not childValue.IsNumber():
      value.Reset()
      return value
    value.SetAsReal(math.sqrt(childValue.GetValue()))
    return value

  def CalculateFloor(self, expression):
    value = expression.GetValue()
    child = expression.GetChild(0)
    childValue = self.CalculateValueOfExpression(child)
    if not childValue.IsNumber():
      value.Reset()
      return value
    value.SetAsReal(math.floor(childValue.GetValue()))
    return value

  def CalculateCeil(self, expression):
    value = expression.GetValue()
    child = expression.GetChild(0)
    childValue = self.CalculateValueOfExpression(child)
    if not childValue.IsNumber():
      value.Reset()
      return value
    value.SetAsReal(math.ceil(childValue.GetValue()))
    return value

  def CalculateAbs(self, expression):
    value = expression.GetValue()
    child = expression.GetChild(0)
    childValue = self.CalculateValueOfExpression(child)
    if not childValue.IsNumber():
      value.Reset()
      return value
    value.SetAsReal(math.fabs(childValue.GetValue()))
    return value

  def CalculateMod(self, expression):
    value = expression.GetValue()
    child1 = expression.GetChild(0)
    childValue1 = self.CalculateValueOfExpression(child1)
    child2 = expression.GetChild(1)
    childValue2 = self.CalculateValueOfExpression(child2)
    if not childValue1.IsNumber() or not childValue2.IsNumber():
      value.Reset()
      return value
    value.SetAsReal(math.fmod(childValue1.GetValue(), childValue2.GetValue()))
    return value

  def CalculateGamma(self, expression):
    value = expression.GetValue()
    child = expression.GetChild(0)
    childValue = self.CalculateValueOfExpression(child)
    if not childValue.IsNumber():
      value.Reset()
      return value
    value.SetAsReal(math.gamma(childValue.GetValue()))
    return value

  def CalculateSin(self, expression):
    value = expression.GetValue()
    child = expression.GetChild(0)
    childValue = self.CalculateValueOfExpression(child)
    if not childValue.IsNumber():
      value.Reset()
      return value
    value.SetAsReal(math.sin(childValue.GetValue()))
    return value

  def CalculateCos(self, expression):
    value = expression.GetValue()
    child = expression.GetChild(0)
    childValue = self.CalculateValueOfExpression(child)
    if not childValue.IsNumber():
      value.Reset()
      return value
    value.SetAsReal(math.cos(childValue.GetValue()))
    return value

  def CalculateTan(self, expression):
    value = expression.GetValue()
    child = expression.GetChild(0)
    childValue = self.CalculateValueOfExpression(child)
    if not childValue.IsNumber():
      value.Reset()
      return value
    value.SetAsReal(math.tan(childValue.GetValue()))
    return value

  def CalculateAsin(self, expression):
    value = expression.GetValue()
    child = expression.GetChild(0)
    childValue = self.CalculateValueOfExpression(child)
    if not childValue.IsNumber():
      value.Reset()
      return value
    value.SetAsReal(math.asin(childValue.GetValue()))
    return value

  def CalculateAcos(self, expression):
    value = expression.GetValue()
    child = expression.GetChild(0)
    childValue = self.CalculateValueOfExpression(child)
    if not childValue.IsNumber():
      value.Reset()
      return value
    value.SetAsReal(math.acos(childValue.GetValue()))
    return value

  def CalculateAtan(self, expression):
    value = expression.GetValue()
    child = expression.GetChild(0)
    childValue = self.CalculateValueOfExpression(child)
    if not childValue.IsNumber():
      value.Reset()
      return value
    value.SetAsReal(math.atan(childValue.GetValue()))
    return value

  def CalculateConstantDistribution(self, expression):
    value = expression.GetValue()
    child = expression.GetChild(0)
    childValue = self.CalculateValueOfExpression(child)
    if not childValue.IsNumber():
      value.Reset()
      return value
    value.Set(childValue)
    return value

  def CalculateExponentialDistribution(self, expression):
    value = expression.GetValue()
    rateChild = expression.GetChild(0)
    rateChildValue = self.CalculateValueOfExpression(rateChild)
    timeChild = expression.GetChild(1)
    timeChildValue = self.CalculateValueOfExpression(timeChild)
    if not rateChildValue.IsNumber() or not timeChildValue.IsNumber():
      value.Reset()
      return value
    rate = rateChildValue.GetValue()
    time = timeChildValue.GetValue()
    value.SetAsReal(1.0-math.exp(-rate*time))
    return value

  def CalculateGLMDistribution(self, expression):
    value = expression.GetValue()
    gammaChild = expression.GetChild(0)
    gammaChildValue = self.CalculateValueOfExpression(gammaChild)
    lambdaChild = expression.GetChild(1)
    lambdaChildValue = self.CalculateValueOfExpression(lambdaChild)
    muChild = expression.GetChild(2)
    muChildValue = self.CalculateValueOfExpression(muChild)
    timeChild = expression.GetChild(3)
    timeChildValue = self.CalculateValueOfExpression(timeChild)
    if not gammaChildValue.IsNumber() or not lambdaChildValue.IsNumber() or not muChildValue.IsNumber() or not timeChildValue.IsNumber():
      value.Reset()
      return value
    gamma = gammaChildValue.GetValue()
    lambda_ = lambdaChildValue.GetValue()
    mu = muChildValue.GetValue()
    time = timeChildValue.GetValue()
    sum = lambda_ + mu
    value.SetAsReal(lambda_/sum + (gamma - lambda_/sum)*math.exp(-sum*time))
    return value

  def CalculateWeibullDistribution(self, expression):
    value = expression.GetValue()
    alphaChild = expression.GetChild(0)
    alphaChildValue = self.CalculateValueOfExpression(alphaChild)
    betaChild = expression.GetChild(1)
    betaChildValue = self.CalculateValueOfExpression(betaChild)
    timeChild = expression.GetChild(2)
    timeChildValue = self.CalculateValueOfExpression(timeChild)
    if not alphaChildValue.IsNumber() or not betaChildValue.IsNumber() or not timeChildValue.IsNumber():
      value.Reset()
      return value
    alpha = alphaChildValue.GetValue()
    beta = betaChildValue.GetValue()
    time = timeChildValue.GetValue()
    value.SetAsReal(1.0-math.exp(-math.pow(time/alpha, beta)))
    return value

  def CalculateDiracDistribution(self, expression):
    value = expression.GetValue()
    delayChild = expression.GetChild(0)
    delayChildValue = self.CalculateValueOfExpression(delayChild)
    timeChild = expression.GetChild(1)
    timeChildValue = self.CalculateValueOfExpression(timeChild)
    if not delayChildValue.IsNumber() or not timeChildValue.IsNumber():
      value.Reset()
      return value
    delay = delayChildValue.GetValue()
    time = timeChildValue.GetValue()
    value.SetAsReal(1.0 if delay<time else 0.0)
    return value

  def CalculatePiecewiseUniformDistribution(self, expression):
    value = expression.GetValue()
    timeChild = expression.GetChild(0)
    timeChildValue = self.CalculateValueOfExpression(timeChild)
    if not timeChildValue.IsNumber():
      value.Reset()
      return value
    time = timeChildValue.GetValue()
    distribution = expression.GetDistribution()
    probability = distribution.GetYInterpolated(time)
    if probability<0:
      probability = 0.0
    if probability>1:
      probability = 1.0
    value.SetAsReal(probability)
    return value

  def CalculateStepwiseDistribution(self, expression):
    value = expression.GetValue()
    timeChild = expression.GetChild(0)
    timeChildValue = self.CalculateValueOfExpression(timeChild)
    if not timeChildValue.IsNumber():
      value.Reset()
      return value
    time = timeChildValue.GetValue()
    distribution = expression.GetDistribution()
    probability = distribution.GetYStepwise(time)
    if probability<0:
      probability = 0.0
    if probability>1:
      probability = 1.0
    value.SetAsReal(probability)
    return value

# 5) Management of random deviates
# --------------------------------

  def ResetAbscissasOfRandomDeviatesInExpression(self, expression):
    if expression.GetType()==S2MLType.REFERENCE_EXPRESSION:
      return
    if expression.GetType()==S2MLType.CONSTANT_EXPRESSION:
      return
    for child in expression.GetChildren():
      self.ResetAbscissasOfRandomDeviatesInExpression(child)
    if expression.GetType()==S2MLType.UNIFORM_DEVIATE_EXPRESSION:
      expression.SetRandomDeviateAbscissa(0.5)
    elif expression.GetType()==S2MLType.NORMAL_DEVIATE_EXPRESSION:
      expression.SetRandomDeviateAbscissa(0.0)
    elif expression.GetType()==S2MLType.LOGNORMAL_DEVIATE_EXPRESSION:
      expression.SetRandomDeviateAbscissa(0.0)
    elif expression.GetType()==S2MLType.EXPONENTIAL_DEVIATE_EXPRESSION:
      expression.SetRandomDeviateAbscissa(0.5)
    elif expression.GetType()==S2MLType.WEIBULL_DEVIATE_EXPRESSION:
      expression.SetRandomDeviateAbscissa(0.5)
    elif expression.GetType()==S2MLType.PIECEWISE_UNIFORM_DEVIATE_EXPRESSION:
      expression.SetRandomDeviateAbscissa(0.5)
    elif expression.GetType()==S2MLType.STEPWISE_DEVIATE_EXPRESSION:
      expression.SetRandomDeviateAbscissa(0.5)

  def DrawAbscissasOfRandomDeviatesInExpression(self, expression):
    if expression.GetType()==S2MLType.REFERENCE_EXPRESSION:
      return
    if expression.GetType()==S2MLType.CONSTANT_EXPRESSION:
      return
    for child in expression.GetChildren():
      self.DrawAbscissasOfRandomDeviatesInExpression(child)
    if expression.GetType()==S2MLType.UNIFORM_DEVIATE_EXPRESSION:
      expression.SetRandomDeviateAbscissa(random.uniform(0.0, 1.0))
    elif expression.GetType()==S2MLType.NORMAL_DEVIATE_EXPRESSION:
      expression.SetRandomDeviateAbscissa(random.normalvariate(0.0, 1.0))
    elif expression.GetType()==S2MLType.LOGNORMAL_DEVIATE_EXPRESSION:
      expression.SetRandomDeviateAbscissa(random.normalvariate(0.0, 1.0))
    elif expression.GetType()==S2MLType.TRIANGULAR_DEVIATE_EXPRESSION:
      expression.SetRandomDeviateAbscissa(random.uniform(0.0, 1.0))
    elif expression.GetType()==S2MLType.EXPONENTIAL_DEVIATE_EXPRESSION:
      expression.SetRandomDeviateAbscissa(random.uniform(0.0, 1.0))
    elif expression.GetType()==S2MLType.WEIBULL_DEVIATE_EXPRESSION:
      expression.SetRandomDeviateAbscissa(random.uniform(0.0, 1.0))
    elif expression.GetType()==S2MLType.PIECEWISE_UNIFORM_DEVIATE_EXPRESSION:
      expression.SetRandomDeviateAbscissa(random.uniform(0.0, 1.0))
    elif expression.GetType()==S2MLType.STEPWISE_DEVIATE_EXPRESSION:
      expression.SetRandomDeviateAbscissa(random.uniform(0.0, 1.0))
